from scipy.integrate import cumulative_trapezoid

import numpy as np
import dcmri as dc

def test__flux_nx():
    n = 10
    Ta = 10
    Fb = 2
    vb = 0.1
    t = np.linspace(0, 20, n)
    ca = np.exp(-t/Ta)/Ta
    J = dc.flux_tissue(ca, t=t, kinetics='NX', vb=vb, Fb=Fb)
    assert J[0] == 0
    Fb = 0
    J = dc.flux_tissue(ca, t=t, kinetics='NX', vb=vb, Fb=Fb)
    assert J[0] == 0


def test__conc_u():
    n = 10
    Ta = 10
    Fb = 2
    t = np.linspace(0, 20, n)
    ca = np.exp(-t/Ta)/Ta
    C = dc.conc_tissue(ca, t=t, kinetics='U', Fb=Fb)
    C0 = Fb*cumulative_trapezoid(ca, t, initial=0)
    assert np.linalg.norm(C-C0)/np.linalg.norm(C0) < 0.01

def test__flux_u():
    n = 10
    Ta = 10
    Fb = 2
    t = np.linspace(0, 20, n)
    ca = np.exp(-t/Ta)/Ta
    J = dc.flux_tissue(ca, t=t, kinetics='U', Fb=Fb)
    J0 = np.zeros(len(t))
    assert np.linalg.norm(J-J0) < 0.01

def test__conc_fx():
    n = 10
    Ta = 10
    Fb = 2
    ve = 0.1
    H = 0.45
    t = np.linspace(0, 20, n)
    ca = np.exp(-t/Ta)/Ta
    C = dc.conc_tissue(ca, t=t, kinetics='FX', H=H, ve=ve, Fb=Fb)
    Fp = Fb*(1-H)
    C0 = Fp*dc.biexpconv(Ta, ve/Fp, t)*ve/Fp/(1-H)
    assert np.linalg.norm(C-C0)/np.linalg.norm(C0) < 0.01
    C = dc.conc_tissue(ca, t=t, kinetics='FX', H=H, ve=ve, Fb=0)
    assert np.linalg.norm(C) == 0

def test__flux_fx():
    n = 10
    Ta = 10
    Fb = 2
    ve = 0.1
    H = 0.45
    t = np.linspace(0, 20, n)
    ca = np.exp(-t/Ta)/Ta
    J = dc.flux_tissue(ca, t=t, kinetics='FX', H=H, ve=ve, Fb=Fb)
    Fp = Fb*(1-H)
    J0 = Fp*dc.biexpconv(Ta, ve/Fp, t)/(1-H)
    assert np.linalg.norm(J-J0)/np.linalg.norm(J0) < 0.01
    J = dc.flux_tissue(ca, t=t, kinetics='FX', H=H, ve=ve, Fb=0)
    assert np.linalg.norm(J) == 0


def test__conc_wv():
    n = 10
    Ta = 10
    Ktrans = 2
    vi = 0.1
    H = 0.45
    t = np.linspace(0, 20, n)
    ca = np.exp(-t/Ta)/Ta
    C = dc.conc_tissue(ca, t=t, kinetics='WV', H=H,  vi=vi, Ktrans=Ktrans)
    C0 = Ktrans*dc.biexpconv(Ta, vi/Ktrans, t)*vi/Ktrans/(1-H)
    assert np.linalg.norm(C-C0)/np.linalg.norm(C0) < 0.01
    C = dc.conc_tissue(ca, t=t, kinetics='WV', H=H, vi=vi, Ktrans=Ktrans)
    assert np.linalg.norm(C-C0)/np.linalg.norm(C0) < 0.01
    C = dc.conc_tissue(ca, t=t, kinetics='WV', H=H, vi=vi, Ktrans=0)
    assert np.linalg.norm(C) == 0


def test__flux_wv():
    n = 10
    Ta = 10
    Ktrans = 2
    kep = 5
    H = 0.45
    vi = Ktrans/kep
    t = np.linspace(0, 20, n)
    ca = np.exp(-t/Ta)/Ta
    J = dc.flux_tissue(ca, t=t, kinetics='WV', H=H, vi=vi, Ktrans=Ktrans)
    J0 = Ktrans*dc.biexpconv(Ta, 1/kep, t)/(1-H)
    assert np.linalg.norm(J[0,1,:]-J0)/np.linalg.norm(J0) < 0.01
    J = dc.flux_tissue(ca, t=t, kinetics='WV', H=H, vi=vi, Ktrans=0)
    assert np.linalg.norm(J[0,1,:]) == 0
    

def test__conc_hfu():
    n = 10
    Ta = 10
    vb = 0.3
    PS = 2
    H = 0.45
    t = np.linspace(0, 20, n)
    ca = np.exp(-t/Ta)/Ta
    C = dc.conc_tissue(ca, t=t, kinetics='HFU', sum=False, H=H, vb=vb, PS=PS)
    C0 = vb*ca
    C1 = PS*dc.conc_trap(ca/(1-H), t)
    assert np.linalg.norm(C[0,:]-C0)/np.linalg.norm(C0) < 0.01
    assert np.linalg.norm(C[1,:]-C1)/np.linalg.norm(C1) < 0.01
    C = dc.conc_tissue(ca, t=t, kinetics='HFU', sum=True, H=H, vb=vb, PS=PS)
    C0 = C0+C1
    assert np.linalg.norm(C-C0)/np.linalg.norm(C0) < 0.01

def test__flux_hfu():
    n = 10
    Ta = 10
    PS = 2
    H = 0.45
    t = np.linspace(0, 20, n)
    ca = np.exp(-t/Ta)/Ta
    J = dc.flux_tissue(ca, kinetics='HFU', H=H, PS=PS)
    J0 = PS*ca/(1-H)
    assert np.linalg.norm(J[1,0,:]-J0)/np.linalg.norm(J0) < 0.01

def test__conc_hf():
    n = 10
    Ta = 10
    vb = 0.3
    PS = 2
    vi = 0.2
    H = 0.45
    t = np.linspace(0, 20, n)
    ca = np.exp(-t/Ta)/Ta
    C = dc.conc_tissue(ca, t=t, kinetics='HF', sum=False, 
                       H=H, vb=vb, vi=vi, PS=PS)
    C0 = vb*ca
    C1 = PS*dc.biexpconv(Ta, vi/PS, t)*vi/PS/(1-H)
    assert np.linalg.norm(C[0,:]-C0)/np.linalg.norm(C0) < 0.01
    assert np.linalg.norm(C[1,:]-C1)/np.linalg.norm(C1) < 0.01
    C = dc.conc_tissue(ca, t=t, kinetics='HF', sum=True, 
                       H=H, vb=vb, vi=vi, PS=PS)
    C0 = C0+C1
    assert np.linalg.norm(C-C0)/np.linalg.norm(C0) < 0.01
    C = dc.conc_tissue(ca, t=t, kinetics='HF', sum=False, 
                       H=H, vb=vb, vi=0, PS=PS)
    C0 = vb*ca
    C1 = np.zeros(len(ca))
    assert np.linalg.norm(C[0,:]-C0)/np.linalg.norm(C0) < 0.01
    assert np.linalg.norm(C[1,:]-C1) == 0
    C = dc.conc_tissue(ca, t=t, kinetics='HF', sum=False, 
                       H=H, vb=vb, vi=vi, PS=0)
    assert 0==np.linalg.norm(C[1,:])

def test__flux_hf():
    n = 10
    Ta = 10
    PS = 2
    kep = 5
    vi = PS/kep
    H = 0.45
    t = np.linspace(0, 20, n)
    ca = np.exp(-t/Ta)/Ta
    J = dc.flux_tissue(ca, t=t, kinetics='HF', H=H, vi=vi, PS=PS)
    J0 = kep*PS*dc.biexpconv(Ta, 1/kep, t)/kep/(1-H)
    assert np.linalg.norm(J[0,1,:]-J0)/np.linalg.norm(J0) < 0.01
    J = dc.flux_tissue(ca, t=t, kinetics='HF', H=H, vi=vi, PS=0)
    assert 0==np.linalg.norm(J[0,1,:])

def test__conc_2cu():
    n = 10
    Ta = 10
    vb = 0.3
    PS = 2
    Fb = 4
    H = 0.45
    Fp = (1-H)*Fb
    vp = (1-H)*vb
    t = np.linspace(0, 20, n)
    ca = np.exp(-t/Ta)/Ta
    C = dc.conc_tissue(ca, t=t, sum=False, kinetics='2CU', 
                       H=H, vb=vb, Fb=Fb, PS=PS)
    Tp = vp/(Fp+PS)
    C0 = Tp*Fp*dc.biexpconv(Ta, Tp, t)/(1-H)
    assert np.linalg.norm(C[0,:]-C0)/np.linalg.norm(C0) < 0.01
    Cs = dc.conc_tissue(ca, t=t, sum=True, kinetics='2CU', 
                        H=H, vb=vb, Fb=Fb, PS=PS)
    C0 = C0+C[1,:]
    assert np.linalg.norm(Cs-C0)/np.linalg.norm(C0) < 0.01
    Cs = dc.conc_tissue(ca, t=t, sum=True, kinetics='2CU', 
                        H=H, vb=0, Fb=Fb, PS=PS)
    C0 = dc.conc_tissue(ca/(1-H), t=t, kinetics='U', 
                        Fb=PS*Fp/(PS+Fp))
    assert np.linalg.norm(Cs-C0)/np.linalg.norm(C0) < 0.01
    Cs = dc.conc_tissue(ca, t=t, sum=True, kinetics='2CU', 
                        H=H, vb=vb, Fb=0, PS=0)
    assert np.linalg.norm(Cs)==0

def test__flux_2cu():
    n = 10
    Ta = 10
    vb = 0.3
    PS = 2
    Fb = 4
    H = 0.45
    vp = (1-H)*vb
    Fp = (1-H)*Fb
    t = np.linspace(0, 20, n)
    ca = np.exp(-t/Ta)/Ta
    J = dc.flux_tissue(ca, t=t, kinetics='2CU', 
                       H=H, vb=vb, Fb=Fb, PS=PS)
    Tp = vp/(Fp+PS)
    J0 = Tp*Fp*dc.biexpconv(Ta, Tp, t)*Fp/vp
    assert np.linalg.norm(J[0,0,:]-J0)/np.linalg.norm(J0) < 0.01
    J = dc.flux_tissue(ca, t=t, kinetics='2CU', 
                       H=H, vb=0, Fb=Fb, PS=PS)
    assert np.linalg.norm(J[0,0,:]-Fb*ca)/np.linalg.norm(Fb*ca) < 0.01

def test__conc_2cx():
    # Compare against general ncomp solution
    t = np.linspace(0, 20, 10)
    J = np.ones((2,len(t)))
    J[1,:] = 0 # No inlet in 2nd compartment
    T = [6,12]
    E = 0.2
    Emat = [[1-E,1],[E,0]]
    Fp = 1/60 # mL/sec/mL
    ca = J[0,:]/Fp
    PS = Fp*E/(1-E)
    vi = T[1]*PS
    vp = T[0]*(Fp+PS)
    H = 0.45
    Fb = Fp/(1-H)
    vb = vp/(1-H)
    C0 = dc.conc_tissue(ca*(1-H), t=t, sum=False, kinetics='2CX', 
                        H=H, vb=vb, vi=vi, Fb=Fb, PS=PS)
    C = dc.conc_ncomp(J, T, Emat, t)
    assert np.linalg.norm(C-C0)/np.linalg.norm(C) < 1e-3
    Cs = dc.conc_tissue(ca*(1-H), t=t, sum=True, kinetics='2CX',
                        H=H, vb=vb, vi=vi, Fb=Fb, PS=PS)
    C0 = np.sum(C,axis=0)
    assert np.linalg.norm(Cs-C0)/np.linalg.norm(C0) < 0.01
    Cs = dc.conc_tissue(ca*(1-H), t=t, sum=True, kinetics='2CX', 
                        H=H, vb=vb, vi=vi, Fb=0, PS=0)
    assert np.linalg.norm(Cs) == 0
    Cs = dc.conc_tissue(ca*(1-H), t=t, sum=False, kinetics='2CX', 
                        H=H, vb=vb, vi=vi, Fb=0, PS=0)
    assert np.linalg.norm(Cs) == 0
    Cs = dc.conc_tissue(ca*(1-H), t=t, sum=True, kinetics='2CX', 
                        H=H, vb=vb, vi=vi, Fb=Fb, PS=0)
    C0 = dc.conc_tissue(ca*(1-H), t=t, kinetics='NX', 
                        vb=vb, Fb=Fb)
    assert np.linalg.norm(Cs-C0)/np.linalg.norm(C) < 1e-3
    C = dc.conc_tissue(ca*(1-H), t=t, sum=False, kinetics='2CX', 
                       H=H, vb=vb, vi=vi, Fb=Fb, PS=0)
    assert np.linalg.norm(C[0,:]-C0)/np.linalg.norm(C0) < 1e-3
    C = dc.conc_tissue(ca*(1-H), t=t, sum=True, kinetics='2CX', 
                       H=H, vb=0, vi=0, Fb=0, PS=0)
    assert np.linalg.norm(C) == 0

    # Test boundaries (Fp=inf)
    Fb = 0.01
    PS = 0.001
    vb = 0.1
    vi = 0.2
    C0 = dc.conc_tissue(ca*(1-H), t=t, sum=False, kinetics='2CX', 
                        H=H, vb=vb, vi=vi, Fb=Fb, PS=PS)
    Fb = 0.05
    C1 = dc.conc_tissue(ca*(1-H), t=t, sum=False, kinetics='2CX', 
                        H=H, vb=vb, vi=vi, Fb=Fb, PS=PS)
    Fb = 0.1
    C2 = dc.conc_tissue(ca*(1-H), t=t, sum=False, kinetics='2CX', 
                        H=H, vb=vb, vi=vi, Fb=Fb, PS=PS)
    Fb = 10.0
    C3 = dc.conc_tissue(ca*(1-H), t=t, sum=False, kinetics='2CX', 
                        H=H, vb=vb, vi=vi, Fb=Fb, PS=PS)
    Fb = np.inf
    C4 = dc.conc_tissue(ca*(1-H), t=t, sum=False, kinetics='2CX', 
                        H=H, vb=vb, vi=vi, Fb=Fb, PS=PS)
    # Check convergence to solution
    err0 = np.linalg.norm(C0[:,1:]-C4[:,1:])
    err1 = np.linalg.norm(C1[:,1:]-C4[:,1:])
    err2 = np.linalg.norm(C2[:,1:]-C4[:,1:])
    err3 = np.linalg.norm(C3[:,1:]-C4[:,1:])
    assert err1 < err0
    assert err2 < err1
    assert err3 < err2
    assert err3 < 1e-2
    # plt.plot(t, C0[0,:], 'r-')
    # plt.plot(t, C0[1,:], 'b-')
    # plt.plot(t, C1[0,:], 'r--')
    # plt.plot(t, C1[1,:], 'b--')
    # plt.plot(t, C2[0,:], 'r-.')
    # plt.plot(t, C2[1,:], 'b-.')
    # plt.plot(t, C3[0,:], 'r-.')
    # plt.plot(t, C3[1,:], 'b-.')
    # plt.plot(t, C4[0,:], 'ro')
    # plt.plot(t, C4[1,:], 'bo')
    # plt.show()

    # Test boundaries (PS+Fp=0)
    Fb = 0
    PS = 0
    vb = 0.1
    vi = 0.2
    C0 = dc.conc_tissue(ca*(1-H), t=t, sum=False, kinetics='2CX', 
                        H=H, vb=vb, vi=vi, Fb=Fb, PS=PS)
    assert np.linalg.norm(C0) == 0

    # Test boundaries (PS=0)
    Fb = 0.01
    PS = 0
    vb = 0.1
    vi = 0.2
    C0 = dc.conc_tissue(ca*(1-H), t=t, sum=True, kinetics='2CX', 
                        H=H, vb=vb, vi=vi, Fb=Fb, PS=PS)
    C1 = dc.conc_tissue(ca*(1-H), t=t, kinetics='NX', 
                        vb=vb, Fb=Fb)
    assert np.linalg.norm(C0-C1) < 1e-9
    
def test__flux_2cx():
    # Compare against general ncomp solution
    t = np.linspace(0, 20, 10)
    J = np.ones((2,len(t)))
    J[1,:] = 0 # No inlet in 2nd compartment
    T = [6,12]
    E = 0.2
    Emat = [[1-E,1],[E,0]]
    Fp = 1/60 # mL/sec/mL
    ca = J[0,:]/Fp
    PS = Fp*E/(1-E)
    vi = T[1]*PS
    vp = T[0]*(Fp+PS)
    H = 0.45
    vb = vp/(1-H)
    Fb = Fp/(1-H) 
    Jo0 = dc.flux_tissue(ca*(1-H), t=t, kinetics='2CX', 
                         H=H, vb=vb, vi=vi, Fb=Fb, PS=PS) 
    Jo = dc.flux_ncomp(J, T, Emat, t)
    assert np.linalg.norm(Jo-Jo0)/np.linalg.norm(Jo) < 1e-1
    Jo0 = dc.flux_tissue(ca*(1-H), t=t, kinetics='2CX', 
                         H=H, vb=vb, vi=vi, Fb=0, PS=0) 
    assert np.linalg.norm(Jo0) == 0
    Jo0 = dc.flux_tissue(ca*(1-H), t=t, kinetics='2CX', 
                         H=H, vb=vb, vi=vi, Fb=Fb, PS=0) 
    Jo = dc.flux_tissue(ca*(1-H), t=t, kinetics='NX', 
                        vb=vb, Fb=Fb)
    assert np.linalg.norm(Jo-Jo0[0,0,:])/np.linalg.norm(Jo) < 1e-3

    Jo = dc.flux_tissue(ca*(1-H), t=t, kinetics='2CX', 
                        H=H, vb=vb, vi=vi, Fb=Fb, PS=0) 
    assert Jo[0,0,0] == 0

    # Test boundary
    Fb = np.inf
    vb = 0.1
    PS = 0.001
    vi = 0.2
    H = 0.45
    J = dc.flux_tissue(ca, t=t, kinetics='2CX', 
                       H=H, vb=vb, vi=vi, Fb=np.inf, PS=PS)
    assert np.isinf(J[0,0,0])



# def test__conc_2cf():
#     # Compare against general ncomp solution
#     t = np.linspace(0, 20, 10)
#     J = np.ones((2,len(t)))
#     J[1,:] = 0 # No inlet in 2nd compartment
#     T = [6,12]
#     E = 0.2
#     Fp = 1/60 # mL/sec/mL
#     ca = J[0,:]/Fp
#     PS = Fp*E/(1-E)
#     Te = T[1]
#     vp = T[0]*(Fp+PS)
#     C0 = dc.conc_tissue(ca, vp, Fp, PS, Te, t=t, sum=False, kinetics='2CF') 
#     Emat = [
#         [1-E, 0],
#         [E  , 1]]
#     C = dc.conc_ncomp(J, T, Emat, t)
#     assert np.linalg.norm(C-C0)/np.linalg.norm(C) < 1e-2
#     Cs = dc.conc_tissue(ca, vp, Fp, PS, Te, t=t, sum=True, kinetics='2CF')
#     C0 = np.sum(C,axis=0)
#     assert np.linalg.norm(Cs-C0)/np.linalg.norm(C0) < 0.01
#     C = dc.conc_tissue(ca, vp, Fp, 0, Te, t=t, sum=False, kinetics='2CF')
#     C0 = dc.conc_comp(J[0,:], vp/Fp, t)
#     assert np.linalg.norm(C[0,:]-C0)/np.linalg.norm(C0) < 1e-12
#     C = dc.conc_tissue(ca, 0, Fp, PS, Te, t=t, sum=False, kinetics='2CF')
#     assert np.linalg.norm(C[0,:]) == 0
#     Cs = dc.conc_tissue(ca, vp, 0, 0, Te, t=t, sum=True, kinetics='2CF')
#     assert np.linalg.norm(Cs) == 0
#     Cs = dc.conc_tissue(ca, vp, 0, 0, Te, t=t, sum=False, kinetics='2CF')
#     assert np.linalg.norm(Cs) == 0


# def test__flux_2cf():
#     # Compare against general ncomp solution
#     t = np.linspace(0, 20, 10)
#     J = np.ones((2,len(t)))
#     J[1,:] = 0 # No inlet in 2nd compartment
#     T = [6,12]
#     E = 0.2
#     Fp = 1/60 # mL/sec/mL
#     ca = J[0,:]/Fp
#     PS = Fp*E/(1-E)
#     Te = T[1]
#     vp = T[0]*(Fp+PS)
#     Jo0 = dc.flux_tissue(ca, vp, Fp, PS, Te, t=t, kinetics='2CF') 
#     Emat = [
#         [1-E, 0],
#         [E  , 1]]
#     Jo = dc.flux_ncomp(J, T, Emat, t)
#     assert np.linalg.norm(Jo-Jo0)/np.linalg.norm(Jo) < 1e-3
#     Jo0 = dc.flux_tissue(ca, vp, 0, 0, Te, t=t, kinetics='2CF') 
#     assert np.linalg.norm(Jo0) == 0


def test_relax_tissue():

    t = np.arange(0, 300, 1.5)
    ca = dc.aif_parker(t, BAT=20)
    R10, r1 = 1/dc.T1(), dc.relaxivity() 
    H = 0.45

    # Test WV limit - exact
    p = {'H':H, 'vb':0.0, 'vi':0.3, 'Fb':0.01, 'PS':0.005, 'PSe':0.01, 'PSc':0.01}
    R1_0, _, _ = dc.relax_tissue(ca*(1-H), R10, r1, t=t, 
                                 kinetics='2CX', water_exchange='RR', **p)
    p = {'H':H, 'vi':0.3, 'Ktrans':0.01*(1-H)*0.005/(0.01*(1-H)+0.005), 'PSc':0.01}
    R1_1, _, _ = dc.relax_tissue(ca*(1-H), R10, r1, t=t, 
                                 kinetics='WV', water_exchange='RR', **p)
    assert np.linalg.norm(R1_0[1:,:]-R1_1) < 1e-9

    # Test WV limit - approx
    p = {'H':H, 'vb':0.5*1e-3, 'vi':0.3, 'Fb':0.01, 'PS':0.005, 'PSe':0.01, 'PSc':0.01}
    R1_0, _, _ = dc.relax_tissue(ca*(1-H), R10, r1, t=t, 
                                 kinetics='2CX', water_exchange='RR', **p)
    p = {'H':H, 'vi':0.3, 'Ktrans':0.01*(1-H)*0.005/(0.01*(1-H)+0.005), 'PSc':0.01}
    R1_1, _, _ = dc.relax_tissue(ca*(1-H), R10, r1, t=t, 
                                 kinetics='WV', water_exchange='RR', **p)
    assert np.linalg.norm(R1_0[1:,:]-R1_1)< 1e-3*np.linalg.norm(R1_0[1:,:])

    # Test HF limit - exact
    p = {'H':H, 'vb':0.05, 'vi':0.3, 'Fb':np.inf, 'PS':0.005, 'PSe':0.01, 'PSc':0.01}
    R1_0, _, _ = dc.relax_tissue(ca, R10, r1, t=t, 
                                 kinetics='2CX', water_exchange='RR', **p)
    p = {'H':H, 'vb':0.05, 'vi':0.3, 'PS':0.005, 'PSe':0.01, 'PSc':0.01}
    R1_1, _, _ = dc.relax_tissue(ca, R10, r1, t=t, 
                                 kinetics='HF', water_exchange='RR', **p)
    assert np.linalg.norm(R1_0-R1_1) < 1e-9

    # Test HF limit - approx
    p = {'H':0.4, 'vb':0.05, 'vi':0.3, 'Fb':1000, 'PS':0.005, 'PSe':0.01, 'PSc':0.01}
    R1_0, _, _ = dc.relax_tissue(ca, R10, r1, t=t, 
                                 kinetics='2CX', water_exchange='RR', **p)
    p = {'H':0.4, 'vb':0.05, 'vi':0.3, 'PS':0.005, 'PSe':0.01, 'PSc':0.01}
    R1_1, _, _ = dc.relax_tissue(ca, R10, r1, t=t, 
                                 kinetics='HF', water_exchange='RR', **p)
    assert np.linalg.norm(R1_0-R1_1) < 1e-3*np.linalg.norm(R1_0)

    # Test NN
    p = {'H':0.4, 'vb':0.05, 'vi':0.3, 'PS':0.005}
    R1_1, _, _ = dc.relax_tissue(ca, R10, r1, t=t, 
                                 kinetics='HF', water_exchange='NN', **p)
    assert 0.6 < R1_1[0,0] < 0.7

    # Run all cases
    for wex in ['FF','RF','FR','RR']:
        for kin in ['U', 'FX', 'NX', 'WV', 'HFU', 'HF', '2CU', '2CX']:
            p = dc.params_tissue(kin, wex)
            p = {key:0.01 for key in p}
            try:
                dc.relax_tissue(ca, R10, r1, t=t, kinetics=kin, 
                                water_exchange=wex, **p)
            except:
                False
            else:
                True
            p = {key:0.0 for key in p}
            try:
                dc.relax_tissue(ca, R10, r1, t=t, kinetics=kin, 
                                water_exchange=wex, **p)
            except:
                False
            else:
                True

    wex = 'RR'
    kin = 'U'
    p = dc.params_tissue(kin, wex)
    p = {key:0 for key in p}
    p['Fb'] = 0.01
    R1, _, _ =dc.relax_tissue(ca, R10, r1, t=t, kinetics=kin, 
                              water_exchange=wex, **p)
    assert 0.6 < R1[0,0] < 0.7

    wex = 'FF'
    kin = '2CU'
    p = dc.params_tissue(kin, wex)
    p = {key:0.01 for key in p}
    p['Fb'] = np.inf
    R1, _, _ =dc.relax_tissue(ca, R10, r1, t=t, kinetics=kin, 
                              water_exchange=wex, **p)
    assert 0.6 < R1[0] < 0.7

    wex = 'FF'
    kin = 'NX'
    p = dc.params_tissue(kin, wex)
    p = {key:0.01 for key in p}
    p['Fb'] = 0
    R1, _, _ =dc.relax_tissue(ca, R10, r1, t=t, kinetics=kin, 
                              water_exchange=wex, **p)
    assert 0.6 < R1[0] < 0.7



    # test exceptions
    try:
        dc.relax_tissue(ca, R10, r1, t=t, kinetics='XX', water_exchange='RR', 
                        **p)
    except:
        assert True
    else:
        assert False
    try:
        dc.relax_tissue(ca, R10, r1, t=t, kinetics='U', water_exchange='XX', 
                        **p)
    except:
        assert True
    else:
        assert False
    try:
        dc.relax_tissue(ca, R10, r1, t=t, kinetics='U', water_exchange='FF')
    except:
        assert True
    else:
        assert False



def test_conc_tissue():
    # Only need to test exceptions
    n = 10
    Ta = 10
    Fb = 2
    t = np.linspace(0, 20, n)
    ca = np.exp(-t/Ta)/Ta
    try:
        dc.conc_tissue(ca, t=t, kinetics='blabla', Fb=Fb)
    except:
        assert True
    else:
        assert False
    

def test_flux_tissue():
    # Only need to test exceptions
    n = 10
    Ta = 10
    Fb = 2
    t = np.linspace(0, 20, n)
    ca = np.exp(-t/Ta)/Ta
    try:
        dc.flux_tissue(ca, t=t, kinetics='blabla', Fb=Fb)
    except:
        assert True
    else:
        assert False


def test_params_tissue():

    try:
        dc.params_tissue('XX', 'RR')
    except:
        assert True
    else:
        assert False

    try:
        dc.params_tissue('2CX', 'XX')
    except:
        assert True
    else:
        assert False

    for kinetics in ['2CX', '2CU', 'HF', 'HFU', 'NX', 'FX', 'WV', 'U']:
        for wxe in ['F','N', 'R']:
            for wxc in ['F','N', 'R']:
                dc.params_tissue(kinetics, wxe+wxc)


def test_signal_tissue():
    nt = 10
    R10 = 1
    r1 = 0.005
    ca = np.ones(nt)
    try:
        dc.signal_tissue(ca, R10, r1, dt=1.0, kinetics='2CX', 
                         water_exchange='FF')
    except:
        assert True
    else:
        assert False
    pars = {
        'H':0.45, 
        'vb': 0.1,
        'vi': 0.3,
        'Fb': 0.5,
        'PS': 0.005,
    }
    seq = {
        'model': 'SS', 
        'S0':1, 
        'FA':15, 
        'TR': 0.001, 
        'B1corr':1,
        'noise_sdev': 0,
    }
    S = dc.signal_tissue(ca, R10, r1, dt=1.0, kinetics='2CX', 
                         water_exchange='FF', sequence=seq, **pars)
    assert 0.007 < S[0] < 0.008
    inflow = {'R10a': 0.7, 'B1corr_a':1}

    try:
        pars_2cu = {
            'H':0.45, 
            'vb': 0.1,
            'Fb': 0.5,
            'PS': 0.005,
        }
        dc.signal_tissue(
            ca, R10, r1, dt=1.0, kinetics='2CU', water_exchange='FF', 
            sequence=seq, inflow=inflow, **pars_2cu)
    except:
        assert True
    else:
        assert False
    S = dc.signal_tissue(
        ca, R10, r1, dt=1.0, kinetics='2CX', water_exchange='FF', 
        sequence=seq, inflow=inflow, **pars)
    assert 0.007 < S[0] < 0.008
    S = dc.signal_tissue(
        ca, R10, r1, dt=1.0, kinetics='2CX', water_exchange='FR', 
        sequence=seq, inflow=inflow, PSc=0.01, **pars)
    assert 0.00735 < S[0] < 0.00736
    
    seq = {
        'model': 'SR', 
        'S0':1, 
        'FA':15, 
        'TR': 0.001, 
        'TC': 0.2,
        'TP': 0.1,
        'B1corr':1,
        'noise_sdev':0,
    }
    S = dc.signal_tissue(ca, R10, r1, dt=1.0, kinetics='2CX', 
                         water_exchange='FF', sequence=seq, **pars)
    assert 0.007 < S[0] < 0.008
    inflow = {'R10a': 0.7, 'B1corr_a':1}
    try:
        pars_2cu = {
            'H':0.45, 
            'vb': 0.1,
            'Fb': 0.5,
            'PS': 0.005,
        }
        dc.signal_tissue(
            ca, R10, r1, dt=1.0, kinetics='2CU', water_exchange='FF', 
            sequence=seq, inflow=inflow, **pars_2cu)
    except:
        assert True
    else:
        assert False
    S = dc.signal_tissue(
        ca, R10, r1, dt=1.0, kinetics='2CX', water_exchange='FF', 
        sequence=seq, inflow=inflow, **pars)
    assert 0.007 < S[0] < 0.008
    S = dc.signal_tissue(
        ca, R10, r1, dt=1.0, kinetics='2CX', water_exchange='FR', 
        sequence=seq, inflow=inflow, PSc=0.01, **pars)
    assert 0.0072 < S[0] < 0.0074


def test_Mz_tissue():
    nt = 10
    R10 = 1
    r1 = 0.005
    ca = np.ones(nt)
    try:
        dc.Mz_tissue(ca, R10, r1, dt=1.0, kinetics='2CX', 
                         water_exchange='FF')
    except:
        assert True
    else:
        assert False
    pars = {
        'H':0.45, 
        'vb': 0.1,
        'vi': 0.3,
        'Fb': 0.5,
        'PS': 0.005,
    }
    seq = {
        'model': 'SS', 
        'S0':1, 
        'FA':15, 
        'TR': 0.001, 
        'B1corr':1,
    }
    Mz = dc.Mz_tissue(ca, R10, r1, dt=1.0, kinetics='2CX', 
                         water_exchange='FF', sequence=seq, **pars)
    assert 0.02 < Mz[0] < 0.03
    inflow = {'R10a': 0.7, 'B1corr_a':1}

    try:
        pars_2cu = {
            'H':0.45, 
            'vb': 0.1,
            'Fb': 0.5,
            'PS': 0.005,
        }
        dc.Mz_tissue(
            ca, R10, r1, dt=1.0, kinetics='2CU', water_exchange='FF', 
            sequence=seq, inflow=inflow, **pars_2cu)
    except:
        assert True
    else:
        assert False
    Mz = dc.Mz_tissue(
        ca, R10, r1, dt=1.0, kinetics='2CX', water_exchange='FF', 
        sequence=seq, inflow=inflow, **pars)
    assert 0.02 < Mz[0] < 0.03
    Mz = dc.Mz_tissue(
        ca, R10, r1, dt=1.0, kinetics='2CX', water_exchange='FR', 
        sequence=seq, inflow=inflow, PSc=0.01, **pars)
    assert 0.01 < Mz[0,0] < 0.02
    
    seq = {
        'model': 'SR', 
        'S0':1, 
        'FA':15, 
        'TR': 0.001, 
        'TC': 0.2,
        'TP': 0.1,
        'B1corr':1,
    }
    Mz = dc.Mz_tissue(ca, R10, r1, dt=1.0, kinetics='2CX', 
                         water_exchange='FF', sequence=seq, **pars)
    assert 0.02 < Mz[0] < 0.03
    inflow = {'R10a': 0.7, 'B1corr_a':1}
    try:
        pars_2cu = {
            'H':0.45, 
            'vb': 0.1,
            'Fb': 0.5,
            'PS': 0.005,
        }
        dc.Mz_tissue(
            ca, R10, r1, dt=1.0, kinetics='2CU', water_exchange='FF', 
            sequence=seq, inflow=inflow, **pars_2cu)
    except:
        assert True
    else:
        assert False
    Mz = dc.Mz_tissue(
        ca, R10, r1, dt=1.0, kinetics='2CX', water_exchange='FF', 
        sequence=seq, inflow=inflow, **pars)
    assert 0.02 < Mz[0] < 0.03
    Mz = dc.Mz_tissue(
        ca, R10, r1, dt=1.0, kinetics='2CX', water_exchange='FR', 
        sequence=seq, inflow=inflow, PSc=0.01, **pars)
    assert 0.01 < Mz[0,0] < 0.02


if __name__ == "__main__":

    test__flux_nx()

    test__conc_u()
    test__flux_u()

    test__conc_fx()
    test__flux_fx()

    test__conc_wv()
    test__flux_wv()

    test__conc_hfu()
    test__flux_hfu()

    test__conc_hf()
    test__flux_hf()

    test__conc_2cu()
    test__flux_2cu()

    test__conc_2cx()
    test__flux_2cx()

    test_conc_tissue()
    test_flux_tissue()

    test_relax_tissue()
    test_Mz_tissue()
    test_signal_tissue()
    test_params_tissue()


    print('All pk_tissue tests passing!')


import math


import numpy as np
from scipy.special import gamma
from scipy.interpolate import CubicSpline
from scipy.integrate import trapezoid




def interp(y, x, pos=False, floor=False) -> np.ndarray:
    """Interpolate uniformly sampled data. 

    This function is a convenience wrapper for standard interpolation, used in dcmri for instance to parametrize non-stationat models.

    Args:
        y (array-like): List of values to interpolate. These are assumed to be uniformly distributed over x.
        x (array-like): Interpolate y at these locations.
        pos (bool, optional): return only positive values. Defaults to False.
        floor (bool, optional): Return only results higher than the lowest value in y. Defaults to False.

    Returns:
        np.ndarray: array of the same length as x, containing the values of y interpolated on x.

    Notes:
        The function uses linear interpolate when y has length 2, quadratic when y has length 3, and cubic spline interpolation otherwise.

    Example:
        >>> import dcmri as dc
        >>> dc.interp([1,3,2], np.arange(10))
        array([1.        , 1.68888889, 2.23333333, 2.63333333, 2.88888889,
               3.        , 2.96666667, 2.78888889, 2.46666667, 2.        ])
    """

    # Interpolate y on x, assuming y-values are uniformly distributed over the x-range
    if np.isscalar(y):
        yi = y*np.ones(len(x))
    elif np.size(y) == 1:
        yi = y[0]*np.ones(len(x))
    elif np.size(y) == 2:
        yi = _lin(x, y)
    elif np.size(y) == 3:
        yi = _quad(x, y)
    else:
        x_y = np.linspace(np.amin(x), np.amax(x), len(y))
        yi = CubicSpline(x_y, y)(x)
    if pos:
        yi[yi < 0] = 0
    if floor:
        y0 = np.amin(y)
        yi[yi < y0] = y0
    return yi


def _quad(t, K):
    # Helper
    nt = len(t)
    mid = math.floor(nt/2)
    return _quadratic(t, t[0], t[mid], t[-1], K[0], K[1], K[2])


def _lin(t, K):
    # Helper
    return _linear(t, t[0], t[-1], K[0], K[1])


def _linear(x, x1, x2, y1, y2):
    # Helper
    # returns a linear function of x
    # that goes through the two points (xi, yi)
    L1 = (x2-x)/(x2-x1)
    L2 = (x-x1)/(x2-x1)
    return y1*L1 + y2*L2


def _quadratic(x, x0, x1, x2, y0, y1, y2):
    # Helper
    # returns a quadratic function of x
    # that goes through the three points (xi, yi)
    L0 = (x-x1)*(x-x2)/((x0-x1)*(x0-x2))
    L1 = (x-x0)*(x-x2)/((x1-x0)*(x1-x2))
    L2 = (x-x0)*(x-x1)/((x2-x0)*(x2-x1))
    return y0*L0 + y1*L1 + y2*L2


def tarray(n, t=None, dt=1.0):
    # Helper function - generate time array.
    if t is None:
        t = dt*np.arange(n)
    else:
        if not isinstance(t, np.ndarray):
            t = np.array(t)
        if len(t) != n:
            raise ValueError('Time array must have same length as the input.')
    return t


def trapz(f, t=None, dt=1.0):
    # Helper function - perform trapezoidal integration.
    # Replace by scipy.integrate.trapezoid
    f = np.array(f)
    n = len(f)
    t = tarray(n, t=t, dt=dt)
    g = np.empty(n)
    g[0] = 0
    for i in range(n-1):
        g[i+1] = g[i] + (t[i+1]-t[i]) * (f[i+1]+f[i]) / 2
    return g


def ddelta(T, t):
    # Helper function - discrete delta
    if not isinstance(t, np.ndarray):
        t = np.array(t)
    n = len(t)
    h = np.zeros(n)
    if T < t[0]:
        return h
    if T > t[-1]:
        return h
    if T == t[0]:
        h[0] = 2/(t[1]-t[0])
        return h
    if T == t[-1]:
        h[-1] = 2/(t[-1]-t[-2])
        return h
    i = np.where(T >= t)[0][-1]
    u = (T-t[i])/(t[i+1]-t[i])
    if i == 0:
        h[i] = (1-u)*2/(t[i+1]-t[i])
    else:
        h[i] = (1-u)*2/(t[i+1]-t[i-1])
    if i == n-2:
        h[i+1] = u*2/(t[i+1]-t[i])
    else:
        h[i+1] = u*2/(t[i+2]-t[i])
    return h


def dstep(T0, T1, t):
    # Helper function - discrete step
    if not isinstance(t, np.ndarray):
        t = np.array(t)
    n = len(t)
    i = np.where((t > T0)*(t < T1))[0]
    if len(i) == 0:
        return ddelta((T0+T1)/2, t)
    i0, i1 = i[0], i[-1]
    t0, t1 = t[i0], t[i1]
    hi = 0
    if i0 > 0:
        u0 = (t0-T0)/(t0-t[i0-1])
        hi += 0.5*(1+u0)*(t0-t[i0-1])
        if i0 > 1:
            hi += 0.5*u0*(t[i0-1]-t[i0-2])
    hi += t1-t0
    if i1 < n-1:
        u1 = (T1-t1)/(t[i1+1]-t1)
        hi += 0.5*(1+u1)*(t[i1+1]-t1)
        if i1 < n-2:
            hi += 0.5*u1*(t[i1+2]-t[i1+1])
    h = np.zeros(n)
    h[i] = 1/hi
    if i0 > 0:
        h[i0-1] = u0/hi
    if i1 < n-1:
        h[i1+1] = u1/hi
    return h


def ddist(H, T, t):
    # discrete distribution - T is an array of times with the boundaries of the histogram bins
    h = np.zeros(len(t))
    for k in range(len(T)-1):
        h += H[k]*dstep(T[k], T[k+1], t)
    return h


def intprod(f, h, t=None, dt=1.0):
    # Helper function
    # Integrate the product of two piecewise linear functions
    # by analytical integration over each interval.
    # Derivation:
    # If f and h are linear between x and y, then we can define slopes:
    # dx = y-x
    # sf(x) = (f(y)-f(x))/dx
    # sh(x) = (h(y)-h(x))/dx
    # With this the integral over the interval becomes:
    # \int_x^y du f(u)h(u)
    # = \int_0^dx du [f(x)+usf(x)] [h(x)+ush(x)]
    # = f(x)*h(x)*dx
    # + [f(x)*sh(x)+sf(x)*h(x)]dx**2/2
    # + sf(x)*sh(x)*dx**3/3
    g = 0
    for l in range(len(f)-1):
        if t is not None:
            dt = t[l+1]-t[l]
        sf = (f[l+1]-f[l])/dt
        sh = (h[l+1]-h[l])/dt
        g += h[l]*f[l]*dt
        g += (sh*f[l]+sf*h[l])*dt**2/2
        g += sh*sf*dt**3/3
    return g


def _intstep(f, h, t=None, dt=1.0):
    g = f*h
    if t is None:
        return 0.5*np.sum(g[1:]+g[:-1])*dt
    else:
        return 0.5*np.sum((g[1:]+g[:-1])*(t[1:]-t[:-1]))


def uconv(f, h, dt=1.0, solver='trap'):
    # Helper function: convolution over uniformly sampled grid.
    n = len(f)
    g = np.zeros(n)
    h = np.flip(h)
    for k in range(1, n):
        if solver == 'trap':
            g[k] = intprod(f[:k+1], h[-(k+1):], dt=dt)
        elif solver == 'step':
            g[k] = _intstep(f[:k+1], h[-(k+1):], dt=dt)
    return g



def conv(f, h, t=None, dt=1.0, solver='step'):
    """Convolve two 1D-arrays.

    This function returns the convolution :math:`f(t)\\otimes h(t)`, using piecewise linear integration to approximate the integrals in the convolution product.

    Args:
        f (array_like): the first 1D array to be convolved.
        h (array_like): the second 1D array to be convolved.
        t (array_like, optional): the time points where the values of f are defined. If t=None, the time points are assumed to be uniformly spaced with spacing dt. Defaults to None.
        dt (float, optional): spacing between time points for uniformly spaced time points. This parameter is ignored if t is explicity provided. Defaults to 1.0.

    Raises:
        ValueError: if f and h have a different length.

    Returns:
        numpy.ndarray: a 1D numpy array of the same length as f and h.

    See Also:
        `expconv`, `biexpconv`, `nexpconv`, `stepconv`

    Notes: 
        The convolution product :math:`f(t)\\otimes h(t)` implemented by `conv` is explicitly defined as:

        .. math::
            g(t) = \\int_0^t du\\, f(u) h(t-u)

        `conv` returns an approximation to this continuous convolution product, calculated by piecewise linear integration. This is not to be confused with other convolution functions, such as `numpy.convolve` which performs discrete convolution. Tracer-kinetic theory is defined by continuous equations and therefore should be performed with `conv` for maximal accuracy, though the difference may be small at high temporal resolution.

        `conv` is generally applicable to any f(t) and h(t), but more accurate formulae for some special cases exists and should be used if available. An example is `expconv`, to be used when either f(t) or h(t) is an exponential.

    Example:
        Import package and create vectors f and h:

        >>> import dcmri as dc
        >>> f = [5,4,3,6]
        >>> h = [2,9,1,3]

        Calculate :math:`g(t) = f(t) \\otimes h(t)` over a uniformly sampled grid of time points with spacing dt=1:

        >>> dc.conv(f, h)
        array([ 0.        , 25.33333333, 41.66666667, 49.        ])

        Calculate the same convolution over a grid of time points with spacing dt=2:

        >>> dc.conv(f, h, dt=2)
        array([ 0.        , 50.66666667, 83.33333333, 98.        ])

        Calculate the same convolution over a non-uniform grid of time points:

        >>> t = [0,1,3,7]
        >>> dc.conv(f, h, t)
        array([  0.        ,  25.33333333,  57.41666667, 108.27083333])
    """
    if len(f) != len(h):
        raise ValueError('f and h must have the same length.')
    if t is None:
        return uconv(f, h, dt, solver=solver)
    n = len(t)
    g = np.zeros(n)
    tf = np.flip(t)
    f = np.flip(f)
    for k in range(1, n):
        tkf = t[k]-tf[-(k+1):]
        tk = np.unique(np.concatenate((t[:k+1], tkf)))
        fk = np.interp(tk, tkf, f[-(k+1):], left=0, right=0)
        hk = np.interp(tk, t[:k+1], h[:k+1], left=0, right=0)
        if solver == 'trap':
            g[k] = intprod(fk, hk, tk)
        elif solver == 'step':
            g[k] = _intstep(fk, hk, tk)
    return g


def inttrap(f, t, t0, t1):
    # Helper function: integrate f from t0 to t1
    ti = t[(t0 < t)*(t < t1)]
    ti = np.concatenate(([t0], ti, [t1]))
    fi = np.interp(ti, t, f, left=0, right=0)
    return trapezoid(fi, ti)


def stepconv(f, T, D, t=None, dt=1.0):
    """Convolve a 1D-array with a normalised step function.

    Args:
        f (array_like): the 1D array to be convolved.
        T (float): the central time point of the step function. 
        D (float): half-width of the step function, as a fraction of T. D must be less or equal to 1.
        t (array_like, optional): the time points where the values of f are defined, in the same units as T. If t=None, the time points are assumed to be uniformly spaced with spacing dt. Defaults to None.
        dt (float, optional): spacing between time points for uniformly spaced time points. This parameter is ignored if t is explicity provided. Defaults to 1.0.

    Raises:
        ValueError: if D > 1.

    Returns:
        numpy.ndarray: a 1D numpy array of the same length as f.

    See Also:
        `conv`, `expconv`, `biexpconv`, `nexpconv`

    Notes: 
        `stepconv` implements the same convolution product as `conv`, but is more accurate and faster in the special case where one of the factors is known to be a step function.

    Example:
        Import package, create a vector f and an array of time points:

        >>> import dcmri as dc
        >>> f = [5,4,3,6]
        >>> t = [0,2,4,7]

        Convolve f with a step function that is centered on t=3 with a half width of 1.5 = 0.5*3: 

        >>> dc.stepconv(f, 3, 0.5, t)
        array([0.        , 0.8125    , 3.64583333, 3.5625    ])
    """
    if D > 1:
        raise ValueError('The dispersion factor D must be <= 1')
    TW = D*T      # Half width of step
    T0 = T-TW     # Initial time point of step
    T1 = T+TW
    n = len(f)
    t = tarray(n, t=t, dt=dt)
    g = np.zeros(n)
    k = len(t[t < T0])
    ti = t[(T0 <= t)*(t <= T1)]
    for tk in ti:
        g[k] = inttrap(f, t, 0, tk-T0)
        k += 1
    ti = t[T1 < t]
    for tk in ti:
        g[k] = inttrap(f, t, tk-T1, tk-T0)
        k += 1
    return g/(2*TW)


def expconv(f, T, t=None, dt=1.0):
    """Convolve a 1D-array with a normalised exponential.

    This function returns the convolution :math:`f(t)\\otimes\\exp(-t/T)/T` using an efficient and accurate numerical formula, as detailed in the appendix of `Flouri et al (2016) <https://onlinelibrary.wiley.com/doi/full/10.1002/mrm.25991>`_ 

    Args:
        f (array_like): the 1D array to be convolved.
        T (float): the characteristic time of the normalized exponential function. 
        t (array_like, optional): the time points where the values of f are defined, in the same units as T. If t=None, the time points are assumed to be uniformly spaced with spacing dt. Defaults to None.
        dt (float, optional): spacing between time points for uniformly spaced time points. This parameter is ignored if t is explicity provided. Defaults to 1.0.

    Returns:
        numpy.ndarray: a 1D numpy array of the same length as f.

    See Also:
        `conv`, `biexpconv`, `nexpconv`, `stepconv`

    Notes: 
        `expconv` implements the same convolution product as `conv`, but is more accurate and faster in the special case where one of the factors is known to be an exponential:

        .. math::
            g(t) = \\frac{e^{-t/T}}{T} \\otimes f(t)

        In code this translates as:

        .. code-block:: python

            g = expconv(f, T, t)

        `expconv` should be used instead of `conv` whenever this applies. Since the transit time distribution of a compartment is exponential, this is an important use case.  

        `expconv` can calculate a convolution between two exponential factors, but in that case an analytical formula can be used which is faster and more accurate. It is implemented in the function `biexpconv`.

    Example:
        Import package and create a vector f:

        >>> import dcmri as dc
        >>> f = [5,4,3,6]

        Calculate :math:`g(t) = f(t) \\otimes \\exp(-t/3)/3` over a uniformly sampled grid of time points with spacing dt=1:

        >>> dc.expconv(f, 3)
        array([0.        , 1.26774952, 1.89266305, 2.6553402 ])

        Calculate the same convolution over a grid of time points with spacing dt=2:

        >>> dc.expconv(f, 3, dt=2)
        array([0.        , 2.16278873, 2.7866186 , 3.70082337])

        Calculate the same convolution over a non-uniform grid of time points:

        >>> t = [0,1,3,7]
        >>> dc.expconv(f, 3, t)
        array([0.        , 1.26774952, 2.32709015, 4.16571645])
    """

    if T == 0:
        return f
    f = np.array(f)
    n = len(f)
    t = tarray(n, t=t, dt=dt)
    x = (t[1:n] - t[0:n-1])/T
    df = (f[1:n] - f[0:n-1])/x
    E = np.exp(-x)
    E0 = 1-E
    E1 = x-E0
    add = f[0:n-1]*E0 + df*E1
    g = np.zeros(n, dtype=E.dtype) 
    for i in range(0, n-1):
        g[i+1] = E[i]*g[i] + add[i]
    return g


def biexpconv(T1, T2, t):
    """Convolve two normalised exponentials analytically.

    Args:
        T1 (float): the characteristic time of the first exponential function. 
        T2 (float): the characteristic time of the second exponential function, in the same units as T1. 
        t (array_like, optional): the time points where the values of f are defined, in the same units as T. 

    Returns:
        numpy.ndarray: The result of the convolution as a 1D array.

    See Also:
        `conv`, `expconv`, `nexpconv`, `stepconv`

    Notes: 
        `biexpconv` returns the exact analytical result of the following convolution:

        .. math::
            g(t) = \\frac{e^{-t/A}}{A} \\otimes \\frac{e^{-t/B}}{B}

        The formula is a biexponential with unit area:

        .. math::
            g(t) = \\frac{Ae^{-t/A}-Be^{-t/B}}{A-B}

        In code this translates as:

        .. code-block:: python

            g = biexpconv(A, B, t)

    Example:
        Import package and create a vector of uniformly sampled time points t with spacing 5.0s:

        >>> import dcmri as dc
        >>> t = 5.0*np.arange(4)

        Calculate the convolution of two normalised exponentials with time constants 10s and 15s:

        >>> g = dc.biexpconv(10, 15, t)
        array([-0.        ,  0.02200013,  0.02910754,  0.02894986])
    """
    if T1 == T2:
        return (t/T1) * np.exp(-t/T1)/T1
    else:
        return (np.exp(-t/T1)-np.exp(-t/T2))/(T1-T2)


def nexpconv(n, T, t):
    """Convolve n identical normalised exponentials analytically

    Args:
        n (float): number of exponentials. Since an analytical formula is used this can also be non-integer.
        T (float): the characteristic time of the exponential. 
        t (array_like, optional): the time points where the values of f are defined, in the same units as T. 

    Returns:
        numpy.ndarray: The result of the convolution as a 1D array.

    Raises:
        ValueError: if n<1 and if T<0

    See Also:
        `conv`, `expconv`, `biexpconv`, `stepconv`

    Notes: 
        `nexpconv` returns the exact analytical result of the following n convolutions:

        .. math::
            g(t) = \\frac{e^{-t/T}}{T} \\otimes \\ldots \\otimes \\frac{e^{-t/T}}{T} 

        The result is a gamma variate function with unit area:

        .. math::
            g(t) = \\frac{1}{\\Gamma(n)}\\left(\\frac{t}{T}\\right)^{n-1} \\frac{e^{-t/T}}{T} 

        Note the gamma-variate function becomes unstable due to numerical 
        overflow at large n and/or short T. 
        This function handles that situation by moving to a numerical 
        solution, but this slows down the computations significantly.
        Callers should if possible avoid this situation by placing suitable 
        bounds on the parameters.

    Example:
        Import package and create a vector of uniformly sampled time points t with spacing 5.0s:

        >>> import dcmri as dc
        >>> t = 5.0*np.arange(4)

        Calculate the convolution of 4 normalised exponentials with time constants 5s:

        >>> g = dc.nexpconv(4, 5, t)
        array([0.        , 0.01226265, 0.03608941, 0.04480836])
    """
    if T < 0:
        raise ValueError('T must be non-negative')
    if n < 1:
        raise ValueError('n cannot be smaller than 1')
    if not isinstance(t, np.ndarray):
        t = np.array(t)
    u = t/T

    # Calculate gamma variate, silencing the warnings
    # as invalid cases are handled properly in the next line.
    with np.errstate(divide='ignore', invalid='ignore', over='ignore', under='ignore'):
        g = u**(n-1) * np.exp(-u)/T/gamma(n)

    # Handle overflow/invalid values
    if False in np.isfinite(g):
        # At large n the analytical formula runs into overflow
        # use numerical calculation in this case (slower and less accurate)
        # The caller is responsible for avoiding this situation by setting suitable parameter bounds
        g = np.exp(-t/T)/T
        n0 = int(np.floor(n))
        for _ in range(n0-1):
            g = expconv(g, T, t)
        if n != n0:
            # Interpolate between n0 and n0+1
            g1 = expconv(g, T, t)
            u = n-n0
            g = g*u + g1*(1-u)
    return g


def sample(t, tp, Sp, dt=None) -> np.ndarray:
    """Sample a signal at given time points.

    Args:
        t (array-like): The time points at which to evaluate the signal.
        tp (array-like): the time points of the signal to be sampled.
        Sp (array-like): the values of the signal to be sampled. Values that are outside of the range are set to zero.
        dt (float, optional): sampling duration. If this is not provided, linear interpolation between the data points is used.  Defaults to None.

    Returns:
        np.ndarray: Signals sampled at times t.
    """
    if dt is None:
        return np.interp(t, tp, Sp, left=0, right=0)
    if dt == 0:
        return np.interp(t, tp, Sp, left=0, right=0)
    Ss = np.zeros(len(t))
    for k, tk in enumerate(t):

        # data = Sp[(tp >= tk) & (tp < tk+dt)]
        # data = Sp[(tp >= tk-dt/2) & (tp < tk+dt/2)]
        # if data.size > 0:
        #     Ss[k] = np.mean(data)

        # NEW (trapezoidal integration - more accurate)
        tb = [tk-dt/2, tk+dt/2]
        Sb = np.interp(tb, tp, Sp)
        i = (tp > tb[0]) & (tp < tb[1])
        ti = np.concatenate(([tb[0]], tp[i], [tb[1]]))
        Si = np.concatenate(([Sb[0]], Sp[i], [Sb[1]]))
        Ss[k] = trapezoid(Si, ti)/dt
    return Ss


def add_noise(signal, sdev: float) -> np.ndarray:
    """Add noise to an MRI magnitude signal.

    Args:
        signal (array-like): Signal values.
        sdev (float): Standard deviation of the noise.

    Returns:
        np.ndarray: signal with noise added.
    """
    noise_x = np.random.normal(0, sdev, np.size(signal))
    noise_y = np.random.normal(0, sdev, np.size(signal))
    signal = np.sqrt((signal+noise_x)**2 + noise_y**2)
    return signal



import numpy as np
from scipy.integrate import trapezoid

import dcmri as dc


def test_tarray():
    n = 4
    J = np.zeros(n)
    t = dc.tarray(len(J))
    assert np.array_equal(t, [0,1,2,3])
    t = dc.tarray(len(J), dt=2)
    assert np.array_equal(t, [0,2,4,6])
    t = dc.tarray(len(J), [1,2,3,9])
    assert np.array_equal(t, [1,2,3,9])
    try:
        t = dc.tarray(len(J), [1,2,3])
    except:
        assert True
    else:
        assert False


def test_trapz():
    t = np.arange(0, 60, 10)
    ca = (t/np.amax(t))**2
    c = dc.trapz(ca, t)
    assert c[1] == 0.20000000000000004


def test_expconv():

    # Uniform time grid with increasing precision: compare against analytical convolution.

    Tf = 20
    Th = 30
    tmax = 30

    prec = [1e-1, 1e-3, 1e-5]
    for i, dt in enumerate([10,1,0.1]):
        t = np.arange(0,tmax,dt)
        f = np.exp(-t/Tf)/Tf
        h = np.exp(-t/Th)/Th
        g = dc.expconv(f, Th, dt=dt)
        g0 = (Tf*f-Th*h)/(Tf-Th)
        assert np.linalg.norm(g-g0)/np.linalg.norm(g0) < prec[i]

    # Non-uniform time grid with increasing precision: check against analytical convolution.
    t0 = np.array([0,1,2,3,5,8,13,21,34])
    prec = [0.05, 1e-2, 1e-3, 1e-5]
    for i, dt0 in enumerate([10,1,0.1,0.01]):
        t = dt0*t0
        f = np.exp(-t/Tf)/Tf
        h = np.exp(-t/Th)/Th
        g = dc.expconv(f, Th, t)
        g0 = (Tf*f-Th*h)/(Tf-Th)
        assert np.linalg.norm(g-g0)/np.linalg.norm(g0) < prec[i]

    #Special case: T=0
    t = np.arange(0,tmax,dt)
    f = np.exp(-t/Tf)/Tf
    assert np.array_equal(f, dc.expconv(f, 0, t))


def test_inttrap():
    t=np.array([0,1,2,3])
    f=[1,1,1,1]
    assert dc.inttrap(f,t,0.5,1.5) == 1

def test_stepconv():
    T = 3.5
    D = 0.5
    T0 = T-D*T
    T1 = T+D*T
    # Check against conv at high res
    prec = [0.04, 0.02, 0.002]
    for k, n in enumerate([10,100,1000]):
        t = np.linspace(0,10,n) 
        h = np.zeros(n)
        h[(t>=T0)*(t<=T1)] = 1/(T1-T0)
        f = np.sqrt(t)
        g = dc.stepconv(f, T, D, dt=t[1])
        g0 = dc.conv(f, h, dt=t[1])
        assert np.linalg.norm(g-g0)/np.linalg.norm(g0) < prec[k]
    try:
        dc.stepconv(f, T, 2, dt=t[1])
    except:
        assert True

def test_intprod():
    # Non-uniform time interval: compare to numerical integration.
    t = [0,2,6]
    f = [1,10,3]
    h = [5,1,7]
    i = dc.intprod(f, h, t)
    n = 1000
    t1 = np.linspace(t[0],t[1],n)
    f1 = np.interp(t1, t[0:2], f[0:2])
    h1 = np.interp(t1, t[0:2], h[0:2])
    i1 = trapezoid(f1*h1, t1)
    t2 = np.linspace(t[1],t[2],n)
    f2 = np.interp(t2, t[1:3], f[1:3])
    h2 = np.interp(t2, t[1:3],h[1:3])
    i2 = trapezoid(f2*h2, t2)
    assert (i-(i1+i2))**2/(i1+i2)**2 < 1e-12

    # Uniform time interval: compare to numerical integration.
    dt = 2
    t = dt*np.arange(3)
    f = [1,10,3]
    h = [5,1,7]
    i = dc.intprod(f, h, dt=dt)
    n = 1000
    t1 = np.linspace(t[0],t[1],n)
    f1 = np.interp(t1, t[0:2], f[0:2])
    h1 = np.interp(t1, t[0:2], h[0:2])
    i1 = trapezoid(f1*h1, t1)
    t2 = np.linspace(t[1],t[2],n)
    f2 = np.interp(t2, t[1:3], f[1:3])
    h2 = np.interp(t2, t[1:3], h[1:3])
    i2 = trapezoid(f2*h2, t2)
    assert (i-(i1+i2))**2/(i1+i2)**2 < 1e-12


def test_uconv():
    # Compare against analytical convolution for 3 time intervals

    Tf = 20
    Th = 30
    tmax = 30

    prec = [1e-1, 1e-3, 1e-5]

    for i, dt in enumerate([10,1,0.1]):
        t = np.arange(0,tmax,dt)
        f = np.exp(-t/Tf)/Tf
        h = np.exp(-t/Th)/Th
        g = dc.uconv(f, h, dt)
        g0 = (Tf*f-Th*h)/(Tf-Th)
        assert np.linalg.norm(g-g0)/np.linalg.norm(g0) < prec[i]


def tfib(n, tmax=1.0):
    t = np.empty(n)
    t[0] = 0
    t[1] = 1
    t[2] = 2
    k=3
    while k<n:
        t[k]=t[k-1]+t[k-2]
        k+=1
    return tmax*t/t.max()

def test_conv():

    # Uniform time grid with increasing precision: compare against analytical convolution.

    Tf = 20
    Th = 30
    tmax = 30

    prec = [1e-1, 1e-3, 1e-5]
    for i, dt in enumerate([10,1,0.1]):
        t = np.arange(0,tmax,dt)
        f = np.exp(-t/Tf)/Tf
        h = np.exp(-t/Th)/Th
        g = dc.conv(f, h, dt=dt)
        g0 = (Tf*f-Th*h)/(Tf-Th)
        assert np.linalg.norm(g-g0)/np.linalg.norm(g0) < prec[i]

    # Non-uniform time grid with increasing precision: check against analytical convolution.
    t0 = np.array([0,1,2,3,5,8,13,21,34])
    prec = [0.05, 0.02, 1e-3, 1e-5]
    for i, dt0 in enumerate([10,1,0.1,0.01]):
        t = dt0*t0
        f = np.exp(-t/Tf)/Tf
        h = np.exp(-t/Th)/Th
        g = dc.conv(f, h, t)
        g0 = (Tf*f-Th*h)/(Tf-Th)
        assert np.linalg.norm(g-g0)/np.linalg.norm(g0) < prec[i]

    # Uniform time grid: check area preserving and symmetric at any time resolution
    nt = [5,10,100]
    tmax = 150
    prec_area = [1e-4, 1e-5, 1e-9]
    prec_symm = 1e-15
    for i, n in enumerate(nt):
        t = np.linspace(0,tmax,n)
        dt = t[1]-t[0]
        f = np.exp(-t/10)
        h = np.exp(-((t-30)/15)**2)
        area = trapezoid(f,t)*trapezoid(h,t)
        g0 = dc.conv(f, h, dt=dt)
        g1 = dc.conv(h, f, dt=dt)
        assert (trapezoid(g0,t)-area)**2/area**2 < 5*prec_area[i]
        assert np.linalg.norm(g0-g1)/np.linalg.norm(g0)  < prec_symm

    # Non-uniform time grid: check area preserving and symmetric
    nt = [5,10,50,100,500]
    tmax = 150
    prec_symm = 1e-14
    prec_area = 0.002
    for i, n in enumerate(nt):
        t = tfib(n, tmax)
        f = np.exp(-t/10)
        h = np.exp(-((t-30)/15)**2)
        area = trapezoid(f,t)*trapezoid(h,t)
        g0 = dc.conv(f, h, t)
        g1 = dc.conv(h, f, t)
        assert (trapezoid(g0,t)-area)**2/area**2 < prec_area
        assert np.linalg.norm(g0-g1)/np.linalg.norm(g0)  < prec_symm

    # compare trap and step solvers - should be identical at high temporal resolution
    dt = 0.1
    t = np.arange(0,tmax,dt)
    f = np.exp(-t/Tf)/Tf
    h = np.exp(-t/Th)/Th
    g0 = (Tf*f-Th*h)/(Tf-Th)
    g = dc.conv(f, h, dt=dt)
    assert np.linalg.norm(g-g0) < 1e-3*np.linalg.norm(g0)
    g = dc.conv(f, h, dt=dt, solver='trap')
    assert np.linalg.norm(g-g0) < 1e-3*np.linalg.norm(g0)
    g = dc.conv(f, h, t)
    assert np.linalg.norm(g-g0) < 1e-3*np.linalg.norm(g0)
    g = dc.conv(f, h, t, solver='trap')
    assert np.linalg.norm(g-g0) < 1e-3*np.linalg.norm(g0)

    # Check error handling
    try:
        dc.conv([1,2,3], [1,2])
    except:
        assert True
    else:
        assert False

def test_biexpconv():
    Tf = 20
    Th = 30
    t = np.array([0,1,2,3,5,8,13,21,34])
    g = dc.biexpconv(Tf, Th, t)
    g0 = (np.exp(-t/Tf)-np.exp(-t/Th))/(Tf-Th)
    assert np.linalg.norm(g-g0) == 0
    g = dc.biexpconv(Th, Th, t)
    g0 = (t/Th) * np.exp(-t/Th)/Th
    assert np.linalg.norm(g-g0) == 0


def test_nexpconv():
    MTT = 20

    # High-res pseudocontinuous
    # Uniform time interval
    t = np.linspace(0, 10*MTT, 1000)

    n=2
    T=MTT/n
    g = dc.nexpconv(n, T, t)
    # Check against analytical
    g0 = (t/T) * np.exp(-t/T)/T
    assert np.linalg.norm(g-g0) == 0
    # Check against expconv
    g0 = dc.expconv(np.exp(-t/T)/T, T, t)
    assert np.linalg.norm(g-g0) < 1e-4
    # Check area = 1
    assert np.abs(trapezoid(g,t)-1) < 1e-3
    # Check MTT
    assert np.abs(trapezoid(t*g,t)-MTT) < 1e-5
    
    n=20
    T=MTT/n
    g = dc.nexpconv(n, T, t)
    # Check area = 1
    assert np.abs(trapezoid(g,t)-1) < 1e-12
    # Check MTT
    assert np.abs(trapezoid(t*g,t)-MTT) < 1e-12

    # In this case the numerical approximation is used
    n=200
    T=MTT/n
    g = dc.nexpconv(n, T, t)
    # Check area = 1
    assert np.abs(trapezoid(g,t)-1) < 1e-12
    # Check MTT
    assert np.abs(trapezoid(t*g,t)-MTT) < 0.1
    # Check case of non-integer n
    g = dc.nexpconv(200.5, T, t)
    assert np.abs(trapezoid(g,t)-1) < 1e-12

    # Test list input format
    g = dc.nexpconv(200.5, T, list(t))
    assert np.abs(trapezoid(g,t)-1) < 1e-12

    # Test exceptions
    try:
        dc.nexpconv(n, -1, t)
    except:
        assert True
    try:
        dc.nexpconv(0.5, T, t)
    except:
        assert True

def test_ddelta():
    t = [0,2,3,4]
    h = dc.ddelta(-1, t)
    assert np.array_equal(h, [0,0,0,0])
    h = dc.ddelta(5, t)
    assert np.array_equal(h, [0,0,0,0])
    h = dc.ddelta(0, t)
    assert np.array_equal(h, [1,0,0,0])
    assert trapezoid(h,t) == 1
    h = dc.ddelta(1, t)
    assert np.abs(trapezoid(h,t)-1) < 1e-12
    h = dc.ddelta(2, t)
    assert np.abs(trapezoid(h,t)-1) < 1e-12
    h = dc.ddelta(3.5, t)
    assert np.abs(trapezoid(h,t)-1) < 1e-12
    h = dc.ddelta(4, t)
    assert np.abs(trapezoid(h,t)-1) < 1e-12

    # Check that this is a unit for the convolution.
    t = tfib(10, 30)
    h = dc.ddelta(0,t)
    f = np.exp(-t/30)/30
    g = dc.conv(f, h, t)
    assert np.linalg.norm(g[1:]-f[1:])/np.linalg.norm(f[1:]) < 1e-2

def test_dstep():
    t = [0,2,3,4]
    h = dc.dstep(0, 4, t)
    assert np.array_equal(h, [0.25,0.25,0.25,0.25])
    assert np.abs(trapezoid(h,t)-1) < 1e-12
    h = dc.dstep(0.5, 3.5, t)
    assert np.abs(trapezoid(h,t)-1) < 1e-12
    t = [0,1,2,3]
    h = dc.dstep(0.5, 2.5, t)
    assert np.array_equal(h, [0.2,0.4,0.4,0.2])
    assert np.abs(trapezoid(h,t)-1) < 1e-12
    t = [-1,0,1,2]
    h = dc.dstep(0.5, 2.5, t)
    assert np.array_equal(h, [0, 0.25, 0.5, 0.5])
    assert np.abs(trapezoid(h,t)-1) < 1e-12
    t = [1,2,3,4]
    h = dc.dstep(0.5, 2.5, t)
    assert np.array_equal(h, [0.5, 0.5, 0.25, 0])
    assert np.abs(trapezoid(h,t)-1) < 1e-12

def test_ddist():
    t = [0,2,3,4]
    h = dc.ddist([1/3,1/3,1/3], [0,2,3,4], t)
    assert np.abs(trapezoid(h,t)-1) < 1e-12
    t = [0,1,2,3]
    h = dc.ddist([1/3,1/3,1/3], [0,1,2,3], t)
    assert np.array_equal(h, [1/3,1/3,1/3,1/3])
    assert np.abs(trapezoid(h,t)-1) < 1e-12
    h = dc.ddist([0.25,0.5,0.25], [0,1,2,3], t)
    assert np.abs(trapezoid(h,t)-1) < 1e-12


def test_interp():
    x = np.arange(3)
    assert np.array_equal(dc.interp(3, x, pos=False, floor=False), [3,3,3])
    assert np.array_equal(dc.interp([3], x, pos=False, floor=False), [3,3,3])
    assert np.array_equal(dc.interp([3,4], x, pos=False, floor=False), [3,3.5,4])
    assert np.array_equal(dc.interp([3,4,5], x, pos=False, floor=False), [3,4,5])
    assert np.array_equal(dc.interp(np.arange(5), x, pos=False, floor=False), [0,2,4])
    assert np.array_equal(dc.interp(np.arange(5), x, pos=True, floor=True), [0,2,4])


def test_sample():

    tp = np.array([2,4,5,7])
    Sp = np.array([1,2,5,9])

    S = dc.sample(np.array([3]), tp, Sp)
    assert np.array_equal(S, [1.5])
    S = dc.sample(np.array([3]), tp, Sp, dt=1)
    assert np.array_equal(S, [1.5])
    S = dc.sample(np.array([3]), tp, Sp, dt=0.1)
    assert np.array_equal(S.astype(np.float32), [1.5])
    S = dc.sample(np.array([3]), tp, Sp, dt=2.0)
    assert np.array_equal(S.astype(np.float32), [1.5])
    S = dc.sample(np.array([3]), tp, Sp, dt=3.0)
    integral = 0.5*1 + 2*1.5 + 0.5*(2+2+0.5*(5-2)/(5-4))/2
    assert np.array_equal(S.astype(np.float32), [integral/3])
    S = dc.sample(np.array([3,6]), tp, Sp, dt=1)
    assert np.array_equal(S, [1.5,7])

def test_add_noise():
    s0 = [1,2,3,4]
    s1 = dc.add_noise(s0, 0)
    assert np.array_equal(s0, s1)




if __name__ == "__main__":

    print('Testing utils..')

    # test_interp()
    # test_tarray()
    # test_trapz()
    # test_ddelta()
    # test_dstep()
    # test_ddist()
    # test_intprod()
    # test_uconv()
    # test_conv()
    # test_inttrap()
    # test_stepconv()
    # test_expconv()
    # test_biexpconv()
    # test_nexpconv()
    # test_sample()
    # test_add_noise()

    print('All utils tests passed!!')



import numpy as np
import dcmri as dc



def test_c_lin():
    R1 = np.arange(10) + 5
    c = dc.c_lin(R1, 0.1)
    assert np.sum(c) == 450.0
    R1 = np.stack((R1, 2*R1))
    c = dc.c_lin(R1, [5,4])
    assert np.sum(c) == 31.5
    c = dc.c_lin(R1, 5)
    assert np.sum(c) == 27


def test_relax():

    # One compartment
    #################
    r1 = 1

    # One time point - ROI
    R10 = 1
    c = 1
    assert dc.relax(c, R10, r1) == 2

    # One time point - image
    shape = (3,3)
    R10 = np.ones(shape)
    c = np.ones(shape)
    assert np.array_equal(dc.relax(c, R10, r1), np.full(shape,2))

    # 4 time points - ROI
    R10 = 1
    c = np.ones(4)
    assert np.array_equal(dc.relax(c, R10, r1), np.full((4,),2))

    # 4 time points - image
    shape = (3,3,4)
    R10 = np.ones(shape[:2])
    c = np.ones(shape)
    assert np.array_equal(dc.relax(c, R10, r1), np.full(shape,2))

    # Two compartments
    ##################
    r1 = [1,1]

    # One time point - ROI
    R10 = [1,1]
    c = [1,1]
    assert np.array_equal(dc.relax(c, R10, r1), np.full((2,),2))

    # One time point - image
    shape = (2,3,3)
    R10 = np.ones(shape)
    c = np.ones(shape)
    assert np.array_equal(dc.relax(c, R10, r1), np.full(shape,2))

    # 4 time points - ROI
    R10 = [1,1]
    c = np.ones((2,4))
    assert np.array_equal(dc.relax(c, R10, r1), np.full((2,4),2))

    # 4 time points - image
    shape = (2,3,3,4)
    R10 = np.ones(shape[:3])
    c = np.ones(shape)
    assert np.array_equal(dc.relax(c, R10, r1), np.full(shape,2))


if __name__ == "__main__":

    #test_c_lin()
    test_relax()

    print('All rel tests passing!')


import matplotlib.pyplot as plt
import numpy as np

import dcmri.pk as pk
import dcmri.kidney as kidney
import dcmri.lib as lib
import dcmri.ui as ui
import dcmri.sig as sig
import dcmri.utils as utils


class Kidney(ui.Model):
    """General model for whole-kidney signals.

    See Also:
        `Liver`, `Tissue`

    Args:
        kinetics (str, optional): Kinetic model for the kidneys. 
          Options are '2CF' (Two-compartment filtration) and 'HF' 
          (High-flow). Defaults to '2CF'. 
        sequence (str, optional): imaging sequence model. Possible 
          values are 'SS' (steady-state), 'SR' (saturation-recovery), 
          and 'lin' (linear). Defaults to 'SS'.
        params (dict, optional): values for the model parameters,
          specified as keyword parameters. Defaults are used for any 
          that are not provided. See table 
          :ref:`Kidney-defaults` for a list of parameters and 
          their default values.

    Notes:

        In the table below, if **Bounds** is None, the parameter is fixed 
        during training. Otherwise it is allowed to vary between the 
        bounds given.

        .. _Kidney-defaults:
        .. list-table:: Kidney parameters. 
            :widths: 15 10 10 10
            :header-rows: 1

            * - Parameter
              - Value
              - Bounds
              - Usage
            * - **General**
              - 
              - 
              - 
            * - field_strength
              - 3
              - None
              - Always
            * - agent
              - 'gadoterate'
              - None
              - Always
            * - t0
              - 0
              - None
              - Always
            * - **Sequence**
              -
              - 
              - 
            * - TS
              - 0
              - None
              - Always
            * - B1corr
              - 1
              - None
              - sequence in ['SS']
            * - FA
              - 15
              - None
              - sequence in ['SR', 'SS']
            * - TR
              - 0.005
              - None
              - sequence in ['SS']
            * - TC
              - 0.1
              - None
              - sequence == 'SR'
            * - TP
              - 0.05
              - None
              - sequence == 'SR'
            * - **AIF**
              - 
              - 
              - 
            * - B1corr_a
              - 1
              - None
              - sequence in ['SS']
            * - R10a
              - 0.7
              - None
              - Always
            * - **Kidney**
              -
              - 
              - 
            * - H
              - 0.45
              - None
              - Always
            * - Ta
              - 0
              - [0, 3]
              - Always
            * - vol
              - 150
              - None
              - Always
            * - Fp
              - 0.02
              - [0, 0.05]
              - kinetics == '2CF'
            * - vp
              - 0.15
              - [0, 0.3]
              - Always
            * - FF
              - 0.1
              - [0, 0.3]
              - kinetics == '2CF'
            * - Tt
              - 120
              - [0, inf]
              - Always
            * - Ft
              - 0.005
              - [0, 0.05]
              - Always
            * - R10
              - 0.65
              - None
              - Always
            * - S0
              - 1.0
              - [0, inf]
              - Always

    Example:

        Use the model to fit minipig data. The AIF is corrupted by 
        inflow effects so for the purpose of this example we will 
        use a standard input function:

    .. plot::
        :include-source:
        :context: close-figs

        >>> import numpy as np
        >>> import pydmr
        >>> import dcmri as dc

        Read the dataset:

        >>> datafile = dc.fetch('minipig_renal_fibrosis')
        >>> data = pydmr.read(datafile, 'nest')
        >>> rois, pars = data['rois']['Pig']['Test'], data['pars']['Pig']['Test']
        >>> time = pars['TS'] * np.arange(len(rois['LeftKidney']))

        Generate an AIF at high temporal resolution (250 msec):

        >>> dt = 0.25
        >>> t = np.arange(0, np.amax(time) + dt, dt) 
        >>> ca = dc.aif_tristan(
        ...    t, 
        ...    agent="gadoterate",
        ...    dose=pars['dose'],
        ...    rate=pars['rate'],
        ...    weight=pars['weight'],
        ...    CO=60,
        ...    BAT=time[np.argmax(rois['Aorta'])] - 20,
        >>> )        

        Initialize the tissue:

        >>> kidney = dc.Kidney(
        ...    ca=ca,
        ...    dt=dt,
        ...    kinetics='HF',
        ...    field_strength=pars['B0'],
        ...    agent="gadoterate",
        ...    t0=pars['TS'] * pars['n0'],
        ...    TS=pars['TS'], 
        ...    TR=pars['TR'],
        ...    FA=pars['FA'],
        ...    R10a=1/dc.T1(pars['B0'], 'blood'),
        ...    R10=1/dc.T1(pars['B0'], 'kidney'),
        >>> )

        Train the kidney on the data:

        >>> kidney.set_free(Ta=[0,30])
        >>> kidney.train(time, rois['LeftKidney'])
        
        Plot the reconstructed signals and concentrations:

        >>> kidney.plot(time, rois['LeftKidney'])

        Print the model parameters:

        >>> kidney.print_params(round_to=4)
        --------------------------------
        Free parameters with their stdev
        --------------------------------
        Arterial mean transit time (Ta): 13.8658 (0.1643) sec
        Plasma volume (vp): 0.0856 (0.003) mL/cm3
        Tubular flow (Ft): 0.0024 (0.0001) mL/sec/cm3
        Tubular mean transit time (Tt): 116.296 (7.6526) sec

    """

    def __init__(
            self, 
            kinetics='2CF', 
            sequence='SS', 
            aif=None, 
            ca=None, 
            t=None, 
            dt=1.0,
            **params,
        ):

        # Check configuration
        if kinetics not in ['2CF', 'HF']:
            raise ValueError(
                f"Kinetic model {kinetics} is not available."
            )
        if sequence not in ['SS', 'SR', 'lin']:
            raise ValueError(
                f"Sequence ' + str(sequence) + ' is not available."
            )
        
        # Config
        self.kinetics = kinetics
        self.sequence = sequence
        
        # Input function
        self.aif = aif
        self.ca = ca
        self.t = t
        self.dt = dt

        # overide defaults
        self._set_defaults(**params)


    def _params(self):
        return PARAMS
    
    def _model_pars(self):

        # General
        pars = ['field_strength', 'agent', 't0']

        # Sequence
        pars += ['TS']
        if self.sequence == 'SR':
            pars += ['B1corr', 'FA', 'TR', 'TC', 'TP']
        elif self.sequence=='SS':
            pars += ['B1corr', 'FA', 'TR']

        # AIF
        if self.aif is not None:
            pars += ['B1corr_a', 'R10a']

        # Kidney
        pars += ['H', 'Ta', 'vol']
        pars += kidney.params_kidney(self.kinetics)
        pars += ['R10', 'S0']
        
        return pars


    def _par_values(self, export=False):

        if export:
            discard = [
                'field_strength', 'agent', 't0', 'TS', 'FA', 'TR', 
                'TC', 'TP', 'H', 'vol', 'R10', 'S0',
            ]
            pars = self._par_values()
            return {p: pars[p] for p in pars.keys() if p not in discard}
        
        pars = self._model_pars()
        p = {par: getattr(self, par) for par in pars}

        if {'Fp', 'H'}.issubset(p):
            p['Fb'] = _div(p['Fp'], 1 - p['H'])
        if {'FF', 'Fp'}.issubset(p):
            p['Ft'] = p['FF']*p['Fp']
        if {'vp', 'Fp', 'Tt'}.issubset(p):
            p['Tp'] = _div(p['vp'], p['Fp']+p['Ft'])
        if {'vp', 'Fp'}.issubset(p):
            p['Tv'] = _div(p['vp'], p['Fp'])
        if {'Ft', 'Fp'}.issubset(p):
            p['E'] = _div(p['Ft'], p['Ft'] + p['Fp'])
        if p['vol'] is not None:
            if {'Ft'}.issubset(p):
                p['GFR'] = p['Ft'] * p['vol']      
            if {'Fp', 'H', 'vol'}.issubset(p):
                p['RBF'] = _div(p['Fp']*p['vol'], 1-p['H'])
                p['RPF'] = p['Fp']*p['vol']

        return p
    

    def time(self):
        """Array of time points.

        Returns:
            np.ndarray: time points in seconds.
        """
        if self.t is None:
            if self.aif is None:
                return self.dt*np.arange(np.size(self.ca))
            else:
                return self.dt*np.arange(np.size(self.aif))
        else:
            return self.t
        
    def _compute_ca(self):

        # Arterial blood concentrations
        if self.ca is None:
            if self.aif is None:
                raise ValueError(
                    "Either aif or ca must be provided \
                    to predict signal data.")
            else:
                r1 = lib.relaxivity(self.field_strength, 'blood', self.agent)
                t = self.time()
                n0 = int(max([np.sum(t < self.t0), 1]))
                if self.sequence == 'SR':
                    self.ca = sig.conc_src(
                        self.aif, self.TC, 1 / self.R10a, r1, n0)
                elif self.sequence == 'SS':
                    self.ca = sig.conc_ss(
                        self.aif, self.TR, self.B1corr_a * self.FA,
                        1 / self.R10a, r1, n0)
                elif self.sequence == 'lin':
                    self.ca = sig.conc_lin(self.aif, 1/self.R10a, r1, n0)


    def conc(self, sum=True):
        """Tissue concentration

        Args:
            sum (bool, optional): If True, this returns the total 
              concentration. Else the function returns the 
              concentration in the individual compartments. Defaults 
              to True.

        Returns:
            numpy.ndarray: Concentration in M
        """

        self._compute_ca()
        ca = pk.flux(
            self.ca, self.Ta, t=self.t, dt=self.dt, model='plug',
        )
        if self.kinetics == '2CF':
            return kidney.conc_kidney(
                ca / (1-self.H), 
                self.Fp, self.vp, self.FF*self.Fp, self.Tt, 
                t=self.t, dt=self.dt, sum=sum, kinetics='2CF',
            )
        elif self.kinetics == 'HF':
            return kidney.conc_kidney(
                ca / (1-self.H), 
                self.vp, self.Ft, self.Tt,
                t=self.t, dt=self.dt, sum=sum, kinetics='HF',
            )

    def relax(self):
        """Longitudinal relaxation rate R1(t).

        Returns:
            np.ndarray: Relaxation rate. Dimensions are (nt,) for a tissue in 
              fast water exchange, or (nc,nt) for a multicompartment tissue 
              outside the fast water exchange limit.
        """
        C = self.conc()
        rb = lib.relaxivity(self.field_strength, 'blood', self.agent)
        return self.R10 + rb*C

    def signal(self) -> np.ndarray:
        """Pseudocontinuous signal S(t) as a function of time.

        Returns:
            np.ndarray: Signal as a 1D array.
        """
        R1 = self.relax()
        if self.sequence == 'SR':
            return sig.signal_spgr(
                self.S0, R1, self.TC, self.TR, 
                self.B1corr * self.FA, self.TP,
            )
        elif self.sequence == 'SS':
            return sig.signal_ss(
                self.S0, R1, self.TR, self.B1corr * self.FA,
            )
        elif self.sequence == 'lin':
            return sig.signal_lin(self.S0, R1)

    def predict(self, xdata):
        t = self.time()
        if np.amax(xdata) > np.amax(t):
            raise ValueError(
                f"The acquisition window is longer than the duration "
                f"of the AIF. The largest time point that can be "
                f"predicted is {np.amax(t)/60} min."
            )
        sig = self.signal()
        return utils.sample(xdata, t, sig, self.TS)

    def train(self, xdata, ydata, **kwargs):
        if self.sequence == 'SR':
            Sref = sig.signal_spgr(
                1, self.R10, self.TR, self.TC, 
                self.B1corr * self.FA, self.TP,
            )
        elif self.sequence == 'SS':
            Sref = sig.signal_ss(
                1, self.R10, self.TR, self.B1corr * self.FA,
            )
        elif self.sequence == 'lin':
            Sref = sig.signal_lin(1, self.R10)
        n0 = int(max([np.sum(xdata < self.t0), 1]))
        self.S0 = np.mean(ydata[:n0]) / Sref
        return ui.train(self, xdata, ydata, **kwargs)

    def plot(self,
             xdata: np.ndarray,
             ydata: np.ndarray,
             ref=None, xlim=None, fname=None, show=True):
        time = self.time()
        if xlim is None:
            xlim = [np.amin(time), np.amax(time)]
        fig, (ax0, ax1) = plt.subplots(1, 2, figsize=(12, 5))
        ax0.set_title('Prediction of the MRI signals.')
        ax0.plot(xdata/60, ydata, marker='o', linestyle='None',
                 color='cornflowerblue', label='Data')
        ax0.plot(time/60, self.predict(time), linestyle='-',
                 linewidth=3.0, color='darkblue', label='Prediction')
        ax0.set(xlabel='Time (min)', ylabel='MRI signal (a.u.)',
                xlim=np.array(xlim)/60)
        ax0.legend()

        C = self.conc(sum=False)
        ax1.set_title('Reconstruction of concentrations')
        if ref is not None:
            ax1.plot(ref['t']/60, 1000*ref['C'], marker='o', linestyle='None',
                     color='cornflowerblue', label='Tissue ground truth')
            ax1.plot(ref['t']/60, 1000*ref['cp'], marker='o', linestyle='None',
                     color='lightcoral', label='Arterial ground truth')
        ax1.plot(time/60, 1000*self.ca, linestyle='-', linewidth=3.0,
                 color='lightcoral', label='Artery')
        # ax1.plot(time/60, 1000*(C[0,:]+C[1,:]), linestyle='-', linewidth=3.0,
        #          color='darkblue', label='Tissue')
        #vb = self.vp / (1-self.H)
        ax1.plot(time/60, 1000*C[0,:], linestyle='-', linewidth=3.0,
                 color='darkred', label='Blood')
        # ax1.plot(time/60, 1000*C[0,:], linestyle='-', linewidth=3.0,
        #          color='darkred', label='Blood')
        ax1.plot(time/60, 1000*C[1,:], linestyle='-', linewidth=3.0,
                 color='darkcyan', label='Tubuli')
        ax1.set(xlabel='Time (min)', ylabel='Concentration (mM)',
                xlim=np.array(xlim)/60)
        ax1.legend()
        if fname is not None:
            plt.savefig(fname=fname)
        if show:
            plt.show()
        else:
            plt.close()

    
PARAMS = {

    # General

    'field_strength': {
        'init': 3,
        'default_free': False,
        'bounds': [0, np.inf],
        'name': 'Magnetic field strength',
        'unit': 'T',
    },
    'agent': {
        'init': 'gadoterate',
        'default_free': False,
        'bounds': None,
        'name': 'Contrast agent',
        'unit': '',
    },
    't0': {
        'init': 0,
        'default_free': False,
        'bounds': [0, np.inf],
        'name': 'Baseline duration',
        'unit': '',
    },

    # Sequence

    'TS': {
        'init': 0,
        'default_free': False,
        'bounds': [0, np.inf],
        'name': 'Sampling time',
        'unit': 'sec',
    },
    'B1corr': {
        'init': 1,
        'default_free': False,
        'bounds': [0, np.inf],
        'name': 'Tissue B1-correction factor',
        'unit': '',
    },
    'FA': {
        'init': 15,
        'default_free': False,
        'bounds': [0, np.inf],
        'name': 'Flip angle',
        'unit': 'deg',
    },
    'TR': {
        'init': 0.005,
        'default_free': False,
        'bounds': [0, np.inf],
        'name': 'Repetition time',
        'unit': 'sec',
    },
    'TC': {
        'init': 0.2,
        'default_free': False,
        'bounds': [0, np.inf],
        'name': 'Time to k-space center',
        'unit': 'sec',
    },
    'TP': {
        'init': 0.05,
        'default_free': False,
        'bounds': [0, np.inf],
        'name': 'Preparation delay',
        'unit': 'sec',
    },

    # AIF

    'B1corr_a': {
        'init': 1,
        'default_free': False,
        'bounds': [0, np.inf],
        'name': 'Arterial B1-correction factor',
        'unit': '',
    },
    'R10a': {
        'init': 1/lib.T1(3.0, 'blood'),
        'default_free': False,
        'bounds': [0, np.inf],
        'name': 'Arterial precontrast R1',
        'unit': 'Hz',
    },

    # Kidney

    'H': {
        'init': 0.45,
        'default_free': False,
        'bounds': [1e-3, 1 - 1e-3],
        'name': 'Tissue Hematocrit',
        'unit': '',
    },
    'Ta': {
        'init': 0,
        'default_free': True,
        'bounds': [0, 3],
        'name': 'Arterial mean transit time',
        'unit': 'sec',
    },
    'vol': {
        'init': None,
        'default_free': False,
        'bounds': None,
        'name': 'Single-kidney volume',
        'unit': 'mL',
    },
    'Fp': {
        'init': 0.02,
        'default_free': True,
        'bounds': [0, 0.05],
        'name': 'Plasma flow',
        'unit': 'mL/sec/cm3',
    },
    'vp': {
        'init': 0.15,
        'default_free': True,
        'bounds': [0, 0.3],
        'name': 'Plasma volume',
        'unit': 'mL/cm3',
    },
    'FF': {
        'init': 0.1,
        'default_free': True,
        'bounds': [0, 0.3],
        'name': 'Filtration fraction',
        'unit': '',
    },
    'Tt': {
        'init': 120,
        'default_free': True,
        'bounds': [0, np.inf],
        'name': 'Tubular mean transit time',
        'unit': 'sec',
    },
    'Ft': {
        'init': 0.005,
        'default_free': True,
        'bounds': [0, 0.05],
        'name': 'Tubular flow',
        'unit': 'mL/sec/cm3',
    },
    'R10': {
        'init': 1/lib.T1(3.0, 'kidney'),
        'default_free': False,
        'bounds': [0, np.inf],
        'name': 'Tissue precontrast R1',
        'unit': 'Hz',
    },
    'S0': {
        'init': 1.0,
        'default_free': False,
        'bounds': [0, np.inf],
        'name': 'Signal scaling factor',
        'unit': 'a.u.',
    },


    # Derived parameters

    'Fb': {
        'name': 'Blood flow',
        'unit': 'mL/sec/cm3',
    },
    'Tv': {
        'name': 'Vascular mean transit time',
        'unit': 'sec',
    },
    'Tp': {
        'name': 'Plasma mean transit time',
        'unit': 'sec',
    },
    'E': {
        'name': 'Extraction fraction',
        'unit': '',
    },
    'GFR': {
        'name': 'Glomerular filtration rate',
        'unit': 'mL/sec',
    },
    'RBF': {
        'name': 'Renal blood flow',
        'unit': 'mL/sec',
    },
    'RPF': {
        'name': 'Renal plasma flow',
        'unit': 'mL/sec',
    },
    'FAcorr': {
        'name': 'B1-corrected Flip Angle',
        'unit': 'deg',
    },

}


def _div(a, b):
    with np.errstate(divide='ignore', invalid='ignore'):
        return np.where(b == 0, 0, np.divide(a, b))



